from pathlib import Path

import numpy as np
import pygame

from overcooked_simulator import ROOT_DIR
from overcooked_simulator.counters import (
    CuttingBoard,
    Trash,
    TomatoDispenser,
    PlateReturn,
    ServingWindow,
)
from overcooked_simulator.game_items import ProgressibleItem, Plate
from overcooked_simulator.game_items import Tomato
from overcooked_simulator.overcooked_environment import Action
from overcooked_simulator.simulation_runner import Simulator

WHITE = (255, 255, 255)
GREY = (190, 190, 190)
BLACK = (0, 0, 0)
COUNTERCOLOR = (240, 240, 240)
LIGHTGREY = (220, 220, 220)
GREEN = (0, 255, 0)
RED = (255, 0, 0)
BLUE = (0, 0, 255)
YELLOW = (255, 255, 0)
BACKGROUND_COLOR = GREY
BACKGROUND_LINES_COLOR = (200, 200, 200)
KNIFE_COLOR = (120, 120, 120)
PLATE_RETURN_COLOR = (170, 170, 240)
BOARD_COLOR = (239, 193, 151)


class PlayerKeyset:
    """Set of keyboard keys for controlling a player.
    First four keys are for movement. Order: Down, Up, Left, Right.
    5th key is for interacting with counters.
    6th key ist for picking up things or dropping them.

    """

    def __init__(self, player_name: str, keys: list[pygame.key]):
        """Creates a player keyset which contains information about which keyboard keys control the player.
        Movement keys in the following order: Down, Up, Left, Right

        Args:
            player_name: The name of the player to control.
            keys: The keys which control this player in the following order: Down, Up, Left, Right, Interact, Pickup.
        """
        self.name = player_name
        self.player_keys = keys
        self.move_vectors = [[-1, 0], [1, 0], [0, -1], [0, 1]]
        self.key_to_movement = {
            key: vec for (key, vec) in zip(self.player_keys[:-2], self.move_vectors)
        }
        self.interact_key = self.player_keys[-2]
        self.pickup_key = self.player_keys[-1]


class PyGameGUI:
    """Visualisation of the overcooked environmnent and reading keyboard inputs using pygame."""

    def __init__(
        self,
        simulator: Simulator,
        player_names: list[str],
        player_keys: list[pygame.key],
    ):
        self.FPS = 60
        self.simulator = simulator
        self.counter_size = self.simulator.env.counter_side_length
        self.window_width, self.window_height = (
            simulator.env.world_width,
            simulator.env.world_height,
        )

        self.player_names = player_names
        self.player_keys = player_keys
        assert len(self.player_names) == len(
            self.player_keys
        ), "Number of players and keysets should match."

        self.player_keysets: list[PlayerKeyset] = [
            PlayerKeyset(player_name, keys)
            for player_name, keys in zip(self.player_names, self.player_keys)
        ]

        self.images_path = Path(ROOT_DIR, "pygame_gui", "images")

    def send_action(self, action: Action):
        """Sends an action to the game environment.

        Args:
            action: The action to be sent. Contains the player, action type and move direction if action is a movement.
        """
        self.simulator.enter_action(action)

    def handle_keys(self):
        """Handles keyboard inputs. Sends action for the respective players. When a key is held down, every frame
        an action is sent in this function.
        """
        keys = pygame.key.get_pressed()
        for player_idx, keyset in enumerate(self.player_keysets):
            relevant_keys = [keys[k] for k in keyset.player_keys]
            if any(relevant_keys[:-2]):
                move_vec = np.zeros(2)
                for idx, pressed in enumerate(relevant_keys[:-2]):
                    if pressed:
                        move_vec += keyset.move_vectors[idx]
                if np.linalg.norm(move_vec) != 0:
                    move_vec = move_vec / np.linalg.norm(move_vec)

                action = Action(f"p{player_idx + 1}", "movement", move_vec)
                self.send_action(action)

    def handle_key_event(self, event):
        """Handles key events for the pickup and interaction keys. Pickup is a single action,
        for interaction keydown and keyup is necessary, because the player has to be able to hold
        the key down.

        Args:
            event: Pygame event for extracting the key action.
        """
        for keyset in self.player_keysets:
            if event.key == keyset.pickup_key and event.type == pygame.KEYDOWN:
                action = Action(keyset.name, "pickup", "pickup")
                self.send_action(action)

            if event.key == keyset.interact_key:
                if event.type == pygame.KEYDOWN:
                    action = Action(keyset.name, "interact", "keydown")
                    self.send_action(action)
                elif event.type == pygame.KEYUP:
                    action = Action(keyset.name, "interact", "keyup")
                    self.send_action(action)

    def draw_background(self):
        """Visualizes a game background."""
        block_size = self.counter_size // 2  # Set the size of the grid block
        for x in range(0, self.window_width, block_size):
            for y in range(0, self.window_height, block_size):
                rect = pygame.Rect(x, y, block_size, block_size)
                pygame.draw.rect(self.screen, BACKGROUND_LINES_COLOR, rect, 1)

    def draw_players(self, state):
        """Visualizes the players as circles with a triangle for the facing diretion.
        If the player holds something in their hands, it is displayed

        Args:
            state: The game state returned by the environment.
        """
        for player in state["players"].values():
            pos = player.pos
            size = player.radius
            color1 = RED if player.name == "p1" else GREEN
            color2 = WHITE

            rect = pygame.Rect(pos[0] - (size / 2), pos[1] - (size / 2), size, size)
            pygame.draw.circle(self.screen, color2, pos, size)
            pygame.draw.rect(self.screen, color1, rect)

            facing = player.facing_direction

            pygame.draw.polygon(
                self.screen,
                BLUE,
                (
                    (pos[0] + (facing[1] * 5), pos[1] - (facing[0] * 5)),
                    (pos[0] - (facing[1] * 5), pos[1] + (facing[0] * 5)),
                    player.pos + (facing * 20),
                ),
            )
            if player.holding is not None:
                holding_item_pos = player.pos + (20 * player.facing_direction)
                self.draw_item(holding_item_pos, player.holding)

    def draw_item(self, pos, item):
        """Visualisation of an item at the specified position. On a counter or in the hands of the player."""
        if isinstance(item, Tomato):
            if item.finished:
                image = pygame.image.load(
                    self.images_path / "tomato_cut.png"
                ).convert_alpha()  # or .convert_alpha()
            else:
                image = pygame.image.load(
                    self.images_path / "tomato.png"
                ).convert_alpha()  # or .convert_alpha()
            rect = image.get_rect()
            rect.center = pos
            self.screen.blit(image, rect)

        if isinstance(item, Plate):
            image = pygame.image.load(
                "overcooked_simulator/pygame_gui/images/plate.png"
            ).convert_alpha()  # or .convert_alpha()
            rect = image.get_rect()
            rect.center = pos
            self.screen.blit(image, rect)

            if item.holds is not None:
                self.draw_item(pos, item.holds)

        if isinstance(item, ProgressibleItem) and not item.finished:
            self.draw_progress_bar(pos, item.progressed_steps, item.steps_needed)

    def draw_progress_bar(self, pos, current, needed):
        if current != 0:
            bar_height = self.counter_size * 0.2
            progress_width = (current / needed) * self.counter_size
            progress_bar = pygame.Rect(
                pos[0] - (self.counter_size / 2),
                pos[1] - (self.counter_size / 2) + self.counter_size - bar_height,
                progress_width,
                bar_height,
            )
            pygame.draw.rect(self.screen, GREEN, progress_bar)

    def draw_counter(self, counter):
        """Visualisation of a counter at its position. If it is occupied by an item, it is also shown.

        Args:
            counter: The counter to visualize.
        """

        counter_rect_outline = pygame.Rect(
            counter.pos[0] - (self.counter_size / 2),
            counter.pos[1] - (self.counter_size / 2),
            self.counter_size,
            self.counter_size,
        )
        pygame.draw.rect(self.screen, COUNTERCOLOR, counter_rect_outline)

        if isinstance(counter, CuttingBoard):
            board_size = 30
            board_rect = pygame.Rect(
                counter.pos[0] - (board_size / 2),
                counter.pos[1] - (board_size / 2),
                board_size,
                board_size,
            )
            pygame.draw.rect(self.screen, BOARD_COLOR, board_rect)

            knife_rect = pygame.Rect(counter.pos[0] + 6, counter.pos[1] - 8, 5, 20)
            pygame.draw.rect(self.screen, KNIFE_COLOR, knife_rect)

        if isinstance(counter, PlateReturn):
            size = 38
            inner = pygame.Rect(
                counter.pos[0] - (size / 2),
                counter.pos[1] - (size / 2),
                size,
                size,
            )
            pygame.draw.rect(self.screen, PLATE_RETURN_COLOR, inner)

        if isinstance(counter, Trash):
            pygame.draw.circle(
                self.screen,
                (80, 80, 80),
                counter.pos,
                self.counter_size * 0.4,
            )
            pygame.draw.circle(self.screen, BLACK, counter.pos, self.counter_size * 0.3)
        if isinstance(counter, TomatoDispenser):
            board_size = 32
            board_rect = pygame.Rect(
                counter.pos[0] - (board_size / 2),
                counter.pos[1] - (board_size / 2),
                board_size,
                board_size,
            )
            pygame.draw.rect(self.screen, RED, board_rect)
            self.draw_item(counter.pos, Tomato())
        if isinstance(counter, ServingWindow):
            board_size = 33
            board_rect = pygame.Rect(
                counter.pos[0] - (board_size / 2),
                counter.pos[1] - (board_size / 2),
                board_size,
                board_size,
            )
            pygame.draw.rect(self.screen, YELLOW, board_rect)

        if counter.occupied_by is not None:
            self.draw_item(counter.pos, counter.occupied_by)

    def draw_counters(self, state):
        """Visualizes the counters in the environment.

        Args:
            state: The game state returned by the environment.
        """
        for counter in state["counters"]:
            self.draw_counter(counter)

    def draw(self, state):
        """Main visualization function.

        Args:
            state: The game state returned by the environment.
        """
        self.screen.fill(BACKGROUND_COLOR)
        self.draw_background()

        self.draw_counters(state)
        self.draw_players(state)

        pygame.display.flip()

    def start_pygame(self):
        """Starts pygame and the gui loop. Each frame the gamestate is visualized and keyboard inputs are read."""
        pygame.init()
        pygame.font.init()

        self.screen = pygame.display.set_mode((self.window_width, self.window_height))
        pygame.display.set_caption("Simple Overcooked Simulator")
        self.screen.fill(BACKGROUND_COLOR)

        clock = pygame.time.Clock()

        # Game loop
        running = True
        while running:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    running = False
                if event.type in [pygame.KEYDOWN, pygame.KEYUP]:
                    self.handle_key_event(event)

            self.handle_keys()
            clock.tick(self.FPS)
            state = self.simulator.get_state()
            self.draw(state)

        pygame.quit()
